@@ -174,6 +174,10 @@ typedef enum _snd_ctl_event_type {
 #define SND_CTL_TLVT_DB_LINEAR		0x0002
 /** TLV type - dB range container */
 #define SND_CTL_TLVT_DB_RANGE		0x0003
+/** TLV type - dB scale specified by min/max values */
+#define SND_CTL_TLVT_DB_MINMAX		0x0004
+/** TLV type - dB scale specified by min/max values (with mute) */
+#define SND_CTL_TLVT_DB_MINMAX_MUTE	0x0005
 
 /** Mute state */
 #define SND_CTL_TLV_DB_GAIN_MUTE	-9999999
